# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree


from typing import (
    IO,
    Any,
    Dict,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    overload,
)

import libcst as cst


def _get_attribute_as_string(attribute: cst.BaseExpression) -> str:
    names = []
    while isinstance(attribute, cst.Attribute):
        # pyre-fixme[16]: `BaseExpression` has no attribute `value`.
        if isinstance(attribute.value.value, cst.Attribute):
            value = _get_attribute_as_string(
                cst.ensure_type(attribute.value, cst.Attribute).value
            )
        else:
            value = _get_name_as_string(attribute.value.value)
        names.append(value)
        attribute = attribute.attr
    if attribute is not None:
        names.append(_get_name_as_string(attribute.value))
    return ".".join(names)


def _get_name_as_string(node: Union[cst.CSTNode, str]) -> str:
    if isinstance(node, cst.Name):
        return node.value
    else:
        # pyre-fixme[7]: Expected `str` but got `Union[CSTNode, str]`.
        return node


class FunctionAnnotation(NamedTuple):
    parameters: cst.Parameters
    returns: Optional[cst.Annotation]


class ImportStatement(NamedTuple):
    module: Union[cst.Name, cst.Attribute]
    names: Set[str]


class ImportCollector(cst.CSTVisitor):
    def __init__(self) -> None:
        self.existing_imports: Set[str] = set()

    def visit_Import(self, node: cst.Import) -> None:
        for imported_name in node.names:
            asname = imported_name.asname
            if asname:
                self.existing_imports.add(_get_attribute_as_string(asname.name))
            else:
                self.existing_imports.add(_get_attribute_as_string(imported_name.name))


class TypeCollector(cst.CSTVisitor):
    def __init__(self, existing_imports: Set[str]) -> None:
        # Qualifier for storing the canonical name of the current function.
        self.qualifier: List[str] = []
        # Store the annotations.
        self.function_annotations: Dict[str, FunctionAnnotation] = {}
        self.attribute_annotations: Dict[str, cst.Annotation] = {}
        self.imports: Dict[str, ImportStatement] = {}
        self.existing_imports: Set[str] = existing_imports

    def visit_ClassDef(self, node: cst.ClassDef) -> None:
        self.qualifier.append(node.name.value)

    def leave_ClassDef(self, node: cst.ClassDef) -> None:
        self.qualifier.pop()

    def visit_FunctionDef(self, node: cst.FunctionDef) -> bool:
        self.qualifier.append(node.name.value)
        if node.returns is not None:
            # pyre-fixme[6]: Expected `CSTNode` for 1st param but got
            #  `Optional[Annotation]`.
            return_annotation = self._create_import_from_annotation(node.returns)
            parameter_annotations = self._import_parameter_annotations(node.params)
            self.function_annotations[".".join(self.qualifier)] = FunctionAnnotation(
                parameters=parameter_annotations, returns=return_annotation
            )
        # pyi files don't support inner functions, return False to stop the traversal.
        return False

    def leave_FunctionDef(self, node: cst.FunctionDef) -> None:
        self.qualifier.pop()

    def visit_AnnAssign(self, node: cst.AnnAssign) -> bool:
        # pyre-fixme[16]: `BaseExpression` has no attribute `value`.
        self.qualifier.append(node.target.value)
        annotation_value = self._create_import_from_annotation(node.annotation)
        self.attribute_annotations[".".join(self.qualifier)] = annotation_value
        return True

    def leave_AnnAssign(self, node: cst.AnnAssign) -> None:
        self.qualifier.pop()

    def visit_ImportFrom(self, node: cst.ImportFrom) -> None:
        module = node.module
        if module is None or isinstance(node, cst.ImportStar):
            return
        # pyre-fixme[6]: Expected `List[CSTNode]` for 1st param but got
        #  `Union[Sequence[ImportAlias], ImportStar]`.
        # pyre-fixme[6]: Expected `str` for 1st param but got `Union[BaseExpression,
        #  str]`.
        self._add_to_imports(node.names, cst.Name(module.value), module.value)

    def _add_annotation_to_imports(
        self, annotation: cst.Attribute
    ) -> Union[cst.Name, cst.Attribute]:
        key = _get_attribute_as_string(annotation.value)
        # Don't attempt to re-import existing imports.
        if key in self.existing_imports:
            return annotation
        self._add_to_imports(
            [cst.ImportAlias(name=annotation.attr)], annotation.value, key
        )
        return annotation.attr

    @overload
    def _handle_Index(self, slice: cst.Index, node: cst.Subscript) -> cst.Subscript:
        pass

    def _handle_Index(  # noqa
        self, slice: cst.Index, node: cst.BaseExpression
    ) -> cst.BaseExpression:
        value = slice.value
        if isinstance(value, cst.Subscript):
            new_slice = slice.with_changes(value=self._handle_Subscript(value))
            return node.with_changes(slice=new_slice)
        elif isinstance(value, cst.Attribute):
            new_slice = slice.with_changes(value=self._add_annotation_to_imports(value))
            return node.with_changes(slice=new_slice)
        else:
            return node

    def _handle_Subscript(self, node: cst.Subscript) -> cst.Subscript:
        slice = node.slice
        if isinstance(slice, list):
            new_slice = []
            for item in slice:
                value = item.slice.value
                if isinstance(value, cst.Attribute):
                    name = self._add_annotation_to_imports(item.slice.value)
                    new_index = item.slice.with_changes(value=name)
                    new_slice.append(item.with_changes(slice=new_index))
                else:
                    if isinstance(item.slice, cst.Index) and not isinstance(
                        item.slice.value, cst.Name
                    ):
                        new_index = item.slice.with_changes(
                            value=self._handle_Index(item.slice, item)
                        )
                        item = item.with_changes(slice=new_index, comma=None)
                    new_slice.append(item)
            return node.with_changes(slice=new_slice)
        elif isinstance(slice, cst.Index):
            return self._handle_Index(slice, node)
        else:
            return node

    def _create_import_from_annotation(self, returns: cst.Annotation) -> cst.Annotation:
        annotation = returns.annotation
        if isinstance(annotation, cst.Attribute):
            attr = self._add_annotation_to_imports(annotation)
            return cst.Annotation(annotation=attr)
        if isinstance(annotation, cst.Subscript):
            value = annotation.value
            if isinstance(value, cst.Name) and value.value == "Type":
                return returns
            return cst.Annotation(annotation=self._handle_Subscript(annotation))
        else:
            return returns

    def _add_to_imports(
        self, names: List[cst.ImportAlias], module: cst.BaseExpression, key: str
    ) -> None:
        names_as_string = [_get_name_as_string(name.name) for name in names]
        set_names = set(names_as_string)
        if key not in self.imports:
            # pyre-fixme[6]: Expected `Union[Attribute, Name]` for 2nd param but got
            #  `BaseExpression`.
            self.imports[key] = ImportStatement(names=set_names, module=module)
        else:
            import_statement = self.imports[key]
            for name in set_names:
                if name not in import_statement.names:
                    import_statement.names.add(name)

    def _import_parameter_annotations(
        self, parameters: cst.Parameters
    ) -> cst.Parameters:
        def update_annotations(parameters: Sequence[cst.Param]) -> List[cst.Param]:
            updated_parameters = []
            for parameter in list(parameters):
                annotation = parameter.annotation
                if annotation is not None:
                    parameter = parameter.with_changes(
                        annotation=self._create_import_from_annotation(annotation)
                    )
                updated_parameters.append(parameter)
            return updated_parameters

        return parameters.with_changes(
            params=update_annotations(parameters.params),
            default_params=update_annotations(parameters.default_params),
        )


class TypeTransformer(cst.CSTTransformer):
    def __init__(
        self,
        function_annotations: Dict[str, FunctionAnnotation],
        attribute_annotations: Dict[str, cst.Annotation],
        imports: Dict[str, ImportStatement],
    ) -> None:
        # Qualifier for storing the canonical name of the current function.
        self.qualifier: List[str] = []
        # Store the annotations.
        self.function_annotations = function_annotations
        self.attribute_annotations = attribute_annotations
        self.toplevel_annotations: Dict[str, cst.CSTNode] = {}
        self.imports = imports
        self.import_statements: List[cst.ImportFrom] = []
        self.is_generated: bool = False

    def _qualifier_name(self) -> str:
        return ".".join(self.qualifier)

    def _annotate_single_target(
        self, node: cst.Assign, updated_node: cst.Assign
    ) -> Union[cst.Assign, cst.AnnAssign]:
        if isinstance(node.targets[0].target, cst.Tuple):
            target = node.targets[0].target
            # pyre-fixme[16]: `BaseAssignTargetExpression` has no attribute `elements`.
            for element in target.elements:
                if not isinstance(element.value, cst.Subscript):
                    name = _get_name_as_string(element.value.value)
                    self._add_to_toplevel_annotations(name)
            return updated_node
        else:
            target = node.targets[0].target
            # pyre-fixme[16]: `BaseAssignTargetExpression` has no attribute `value`.
            name = _get_name_as_string(target.value)
            self.qualifier.append(name)
            if self._qualifier_name() in self.attribute_annotations and not isinstance(
                target, cst.Subscript
            ):
                annotation = self.attribute_annotations[self._qualifier_name()]
                self.qualifier.pop()
                return cst.AnnAssign(cst.Name(name), annotation, node.value)
            else:
                self.qualifier.pop()
                return updated_node

    def _split_module(
        self, module: cst.Module, updated_module: cst.Module
    ) -> Tuple[
        List[Union[cst.SimpleStatementLine, cst.BaseCompoundStatement]],
        List[Union[cst.SimpleStatementLine, cst.BaseCompoundStatement]],
    ]:
        import_add_location = 0
        # This works under the principle that while we might modify node contents,
        # we have yet to modify the number of statements. So we can match on the
        # original tree but break up the statements of the modified tree. If we
        # change this assumption in this visitor, we will have to change this code.
        for i, statement in enumerate(module.body):
            if isinstance(statement, cst.SimpleStatementLine):
                for possible_import in statement.body:
                    for last_import in self.import_statements:
                        if possible_import is last_import:
                            import_add_location = i + 1
                            break

        return (
            list(updated_module.body[:import_add_location]),
            list(updated_module.body[import_add_location:]),
        )

    def _add_to_toplevel_annotations(self, name: str) -> None:
        self.qualifier.append(name)
        if self._qualifier_name() in self.attribute_annotations:
            annotation = self.attribute_annotations[self._qualifier_name()]
            self.toplevel_annotations[name] = annotation
        self.qualifier.pop()

    def _update_parameters(
        self, annotations: FunctionAnnotation, updated_node: cst.FunctionDef
    ) -> cst.Parameters:
        # Update params and default params with annotations
        # don't override existing annotations or default values
        def update_annotation(
            parameters: Sequence[cst.Param], annotations: Sequence[cst.Param]
        ) -> List[cst.Param]:
            parameter_annotations = {}
            annotated_parameters = []
            for parameter in list(annotations):
                if parameter.annotation:
                    parameter_annotations[parameter.name.value] = parameter.annotation
            for parameter in list(parameters):
                key = parameter.name.value
                if key in parameter_annotations and not parameter.annotation:
                    parameter = parameter.with_changes(
                        annotation=parameter_annotations[key]
                    )
                annotated_parameters.append(parameter)
            return annotated_parameters

        return annotations.parameters.with_changes(
            default_params=update_annotation(
                updated_node.params.default_params,
                annotations.parameters.default_params,
            ),
            params=update_annotation(
                updated_node.params.params, annotations.parameters.params
            ),
        )

    def _insert_empty_line(
        self,
        statements: List[Union[cst.SimpleStatementLine, cst.BaseCompoundStatement]],
    ) -> List[Union[cst.SimpleStatementLine, cst.BaseCompoundStatement]]:
        if len(statements) < 1:
            # No statements, nothing to add to
            return statements
        if len(statements[0].leading_lines) == 0:
            # Statement has no leading lines, add one!
            return [
                statements[0].with_changes(leading_lines=(cst.EmptyLine(),)),
                *statements[1:],
            ]
        if statements[0].leading_lines[0].comment is None:
            # First line is empty, so its safe to leave as-is
            return statements
        # Statement has a comment first line, so lets add one more empty line
        return [
            statements[0].with_changes(
                leading_lines=(cst.EmptyLine(), *statements[0].leading_lines)
            ),
            *statements[1:],
        ]

    def visit_ClassDef(self, node: cst.ClassDef) -> None:
        self.qualifier.append(node.name.value)

    def leave_ClassDef(
        self, original_node: cst.ClassDef, updated_node: cst.ClassDef
    ) -> cst.ClassDef:
        self.qualifier.pop()
        return updated_node

    def visit_FunctionDef(self, node: cst.FunctionDef) -> bool:
        self.qualifier.append(node.name.value)
        # pyi files don't support inner functions, return False to stop the traversal.
        return False

    def leave_FunctionDef(
        self, original_node: cst.FunctionDef, updated_node: cst.FunctionDef
    ) -> cst.FunctionDef:
        key = self._qualifier_name()
        self.qualifier.pop()
        if key in self.function_annotations:
            annotations = self.function_annotations[key]
            # Only add new annotation if one doesn't already exist
            if not updated_node.returns:
                updated_node = updated_node.with_changes(returns=annotations.returns)
            # Don't override default values when annotating functions
            new_parameters = self._update_parameters(annotations, updated_node)
            return updated_node.with_changes(params=new_parameters)
        return updated_node

    def leave_Assign(
        self, original_node: cst.Assign, updated_node: cst.Assign
    ) -> Union[cst.Assign, cst.AnnAssign]:

        if len(original_node.targets) > 1:
            for assign in original_node.targets:
                if not isinstance(assign.target, cst.Subscript):
                    self._add_to_toplevel_annotations(
                        # pyre-fixme[16]: `BaseAssignTargetExpression` has no
                        #  attribute `value`.
                        _get_name_as_string(assign.target.value)
                    )
            return updated_node
        else:
            return self._annotate_single_target(original_node, updated_node)

    def leave_ImportFrom(
        self, original_node: cst.ImportFrom, updated_node: cst.ImportFrom
    ) -> cst.ImportFrom:
        self.import_statements.append(original_node)
        # pyre-fixme[6]: Expected `Union[Attribute, Name]` for 1st param but got
        #  `Optional[Union[Attribute, Name]]`.
        key = _get_attribute_as_string(original_node.module)
        import_names = updated_node.names
        module = original_node.module
        if (
            module is not None
            and module.value in self.imports
            and not isinstance(import_names, cst.ImportStar)
        ):
            names_as_string = [_get_name_as_string(name.name) for name in import_names]
            updated_names = self.imports[key].names.union(set(names_as_string))
            names = [cst.ImportAlias(cst.Name(name)) for name in sorted(updated_names)]
            updated_node = updated_node.with_changes(names=tuple(names))
            del self.imports[key]
        return updated_node

    def visit_Comment(self, node: cst.Comment) -> None:
        if "@" "generated" in node.value:
            self.is_generated = True

    def leave_Module(
        self, original_node: cst.Module, updated_node: cst.Module
    ) -> cst.Module:
        if self.is_generated:
            return original_node
        if not self.toplevel_annotations and not self.imports:
            return updated_node
        toplevel_statements = []
        # First, find the insertion point for imports
        statements_before_imports, statements_after_imports = self._split_module(
            original_node, updated_node
        )

        # Make sure there's at least one empty line before the first non-import
        statements_after_imports = self._insert_empty_line(statements_after_imports)

        imported = set()
        for statement in self.import_statements:
            names = statement.names
            if isinstance(names, cst.ImportStar):
                continue
            for name in names:
                if name.asname:
                    name = name.asname
                if name:
                    imported.add(_get_name_as_string(name.name))

        for _, import_statement in self.imports.items():
            # Filter out anything that has already been imported.
            names = import_statement.names.difference(imported)
            names = [cst.ImportAlias(cst.Name(name)) for name in sorted(names)]
            if not names:
                continue
            import_statement = cst.ImportFrom(
                module=import_statement.module, names=names
            )
            # Add import statements to module body.
            # Need to assign an Iterable, and the argument to SimpleStatementLine
            # must be subscriptable.
            toplevel_statements.append(cst.SimpleStatementLine([import_statement]))

        for name, annotation in self.toplevel_annotations.items():
            annotated_assign = cst.AnnAssign(
                cst.Name(name),
                # pyre-fixme[16]: `CSTNode` has no attribute `annotation`.
                cst.Annotation(annotation.annotation),
                None,
            )
            toplevel_statements.append(cst.SimpleStatementLine([annotated_assign]))

        return updated_node.with_changes(
            body=[
                *statements_before_imports,
                *toplevel_statements,
                *statements_after_imports,
            ]
        )


def _parse(file: IO[Any]) -> cst.Module:  # pyre-fixme[2]
    contents = file.read()
    return cst.parse_module(contents)


def _annotate_source(stubs: cst.Module, source: cst.Module) -> cst.Module:
    import_visitor = ImportCollector()
    source.visit(import_visitor)
    visitor = TypeCollector(import_visitor.existing_imports)
    stubs.visit(visitor)
    transformer = TypeTransformer(
        visitor.function_annotations, visitor.attribute_annotations, visitor.imports
    )
    return source.visit(transformer)


def apply_stub_annotations(stub_path: str, file_path: str) -> str:
    with open(stub_path) as stub_file, open(file_path) as source_file:
        stubs = _parse(stub_file)
        source = _parse(source_file)
        modified_tree = _annotate_source(stubs, source)
        return modified_tree.code



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import argparse
import logging
import os
import shutil
import sys
import time
import traceback
from typing import Type  # noqa

from . import (
    buck,
    commands,
    find_project_root,
    get_binary_version_from_file,
    log,
    log_statistics,
)
from .commands import (  # noqa
    Command,
    ExitCode,
    IncrementalStyle,
    ProfileOutput,
    reporting,
)
from .exceptions import EnvironmentException
from .version import __version__


LOG = logging.getLogger(__name__)  # type: logging.Logger


def main() -> int:
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        formatter_class=argparse.RawTextHelpFormatter,
        epilog="environment variables:"
        "\n   `PYRE_BINARY` overrides the pyre binary used."
        "\n   `PYRE_VERSION_HASH` overrides the pyre version set in the "
        "configuration files.",
    )
    commands.Command.add_arguments(parser)

    # Subcommands.
    subcommand_names = ", ".join([command.NAME for command in commands.COMMANDS])
    parsed_commands = parser.add_subparsers(
        metavar="{}".format(subcommand_names),
        help="""
        The pyre command to run; defaults to `incremental`.
        Run `pyre command --help` for documentation on a specific command.
        """,
    )

    for command in commands.COMMANDS:
        command.add_subparser(parsed_commands)

    arguments = parser.parse_args()

    if not hasattr(arguments, "command"):
        if shutil.which("watchman"):
            # pyre-fixme[16]: `Namespace` has no attribute `command`.
            arguments.command = commands.Incremental
            # pyre-fixme[16]: `Namespace` has no attribute `nonblocking`.
            arguments.nonblocking = False
            # pyre-fixme[16]: `Namespace` has no attribute `incremental_style`.
            arguments.incremental_style = None
            # pyre-fixme[16]: `Namespace` has no attribute `no_start`.
            arguments.no_start = False
        else:
            watchman_link = "https://facebook.github.io/watchman/docs/install.html"
            LOG.warning(
                "No watchman binary found. \n"
                "To enable pyre incremental, "
                "you can install watchman: {}".format(watchman_link)
            )
            LOG.warning("Defaulting to non-incremental check.")
            arguments.command = commands.Check

    command = None
    # Having this as a fails-by-default helps flag unexpected exit
    # from exception flows.
    exit_code = ExitCode.FAILURE
    start = time.time()
    try:
        if arguments.version:
            binary_version = get_binary_version_from_file(arguments.local_configuration)
            log.stdout.write(
                "Binary version: {}\nClient version: {}".format(
                    binary_version, __version__
                )
            )
            return ExitCode.SUCCESS
        original_directory = os.getcwd()
        # TODO(T57959968): Stop changing the directory in the client
        os.chdir(find_project_root(original_directory))
        command = arguments.command(arguments, original_directory)

        log.initialize(command.noninteractive, command.log_directory)
        exit_code = command.run().exit_code()
    except buck.BuckException as error:
        LOG.error(str(error))
        if arguments.command == commands.Persistent:
            commands.Persistent.run_null_server(timeout=3600 * 12)
        exit_code = ExitCode.BUCK_ERROR
    except EnvironmentException as error:
        LOG.error(str(error))
        if arguments.command == commands.Persistent:
            commands.Persistent.run_null_server(timeout=3600 * 12)
        exit_code = ExitCode.FAILURE
    except commands.ClientException as error:
        LOG.error(str(error))
        exit_code = ExitCode.FAILURE
    except Exception as error:
        LOG.error(str(error))
        LOG.info(traceback.format_exc())
        exit_code = ExitCode.FAILURE
    except KeyboardInterrupt:
        LOG.warning("Interrupted by user")
        LOG.debug(traceback.format_exc())
        exit_code = ExitCode.SUCCESS
    finally:
        log.cleanup()
        if command and isinstance(command, Command):
            command.analysis_directory.cleanup()
            configuration = command._configuration
            if configuration and configuration.logger:
                log_statistics(
                    "perfpipe_pyre_usage",
                    arguments=arguments,
                    configuration=configuration,
                    integers={
                        "exit_code": exit_code,
                        "runtime": int((time.time() - start) * 1000),
                    },
                    normals={
                        "cwd": os.getcwd(),
                        "client_version": __version__,
                        "command": command.NAME,
                    },
                )

    return exit_code


if __name__ == "__main__":
    try:
        os.getcwd()
    except FileNotFoundError:
        LOG.error(
            "Pyre could not determine the current working directory. "
            "Has it been removed?\nExiting."
        )
        sys.exit(ExitCode.FAILURE)
    sys.exit(main())



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import argparse
import fnmatch
import json
import logging
import os
from typing import Any, Dict, List, Optional, Sequence, Set  # noqa

from .. import log
from ..analysis_directory import AnalysisDirectory
from ..configuration import Configuration
from ..error import Error
from ..filesystem import translate_path
from .command import TEXT, ClientException, Command, Result


LOG = logging.getLogger(__name__)  # type: logging.Logger


class Reporting(Command):
    NAME = "reporting"

    def __init__(
        self,
        arguments: argparse.Namespace,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super().__init__(
            arguments, original_directory, configuration, analysis_directory
        )

    def _print(self, errors: Sequence[Error]) -> None:
        if errors:
            length = len(errors)
            LOG.error("Found %d type error%s!", length, "s" if length > 1 else "")
        else:
            LOG.log(log.SUCCESS, "No type errors found")

        if self._output == TEXT:
            log.stdout.write("\n".join([repr(error) for error in errors]))
        else:
            log.stdout.write(json.dumps([error.__dict__ for error in errors]))

    def _get_directories_to_analyze(self) -> Set[str]:
        current_project_directories = self._analysis_directory.get_filter_root()
        # The server may not exist in the same directory, so use absolute paths.
        directories_to_analyze = {
            translate_path(os.getcwd(), filter_root)
            for filter_root in current_project_directories
        }
        return directories_to_analyze

    def _get_errors(
        self, result: Result, bypass_filtering: bool = False
    ) -> Sequence[Error]:
        result.check()

        errors = []  # type: List[Error]
        # pyre-ignore: T39175181
        results = {}  # type: List[Dict[str, Any]]
        try:
            results = json.loads(result.output)
            # TODO(T39755668): deprecate 'if' condition eventually.
            if "errors" in results:
                results = results["errors"]
        except (json.JSONDecodeError, ValueError):
            raise ClientException("Invalid output: `{}`.".format(result.output))

        for error in results:
            full_path = os.path.realpath(
                os.path.join(self._analysis_directory.get_root(), error["path"])
            )
            # Relativize path to user's cwd.
            relative_path = self._relative_path(full_path)
            error["path"] = relative_path
            ignore_error = False
            external_to_global_root = True
            if full_path.startswith(self._current_directory):
                external_to_global_root = False
            if not os.path.exists(full_path):
                # Nonexistent paths can be created when search path stubs are renamed.
                external_to_global_root = True
            for absolute_ignore_path in self._ignore_all_errors_paths:
                if fnmatch.fnmatch(full_path, (absolute_ignore_path + "*")):
                    ignore_error = True
                    break
            errors.append(Error(ignore_error, external_to_global_root, **error))

        if bypass_filtering:
            return errors
        else:
            filtered_errors = [
                error
                for error in errors
                if (
                    not error.is_ignored()
                    and (self._verbose or not (error.is_external_to_global_root()))
                )
            ]
            sorted_errors = sorted(
                filtered_errors,
                key=lambda error: (error.path, error.line, error.column),
            )

            return sorted_errors



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import argparse
from typing import Optional

from ..analysis_directory import AnalysisDirectory, resolve_analysis_directory
from ..configuration import Configuration
from .command import Command, IncrementalStyle
from .incremental import Incremental
from .start import Start
from .stop import Stop


class Restart(Command):
    NAME = "restart"

    def __init__(
        self,
        arguments,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super(Restart, self).__init__(
            arguments, original_directory, configuration, analysis_directory
        )
        self._terminal: bool = arguments.terminal
        self._store_type_check_resolution: bool = arguments.store_type_check_resolution
        self._use_watchman: bool = not arguments.no_watchman
        self._incremental_style: IncrementalStyle = arguments.incremental_style

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        restart = parser.add_parser(
            cls.NAME,
            epilog="Restarts a server. Equivalent to `pyre stop && pyre start`.",
        )
        restart.set_defaults(command=cls)
        restart.add_argument(
            "--terminal", action="store_true", help="Run the server in the terminal."
        )
        restart.add_argument(
            "--store-type-check-resolution",
            action="store_true",
            help="Store extra information for `types` queries.",
        )
        restart.add_argument(
            "--no-watchman",
            action="store_true",
            help="Do not spawn a watchman client in the background.",
        )
        restart.add_argument(
            "--incremental-style",
            type=IncrementalStyle,
            choices=list(IncrementalStyle),
            default=None,
            help="How to approach doing incremental checks.",
        )

    def generate_analysis_directory(self) -> AnalysisDirectory:
        return resolve_analysis_directory(
            self._arguments,
            self._configuration,
            self._original_directory,
            self._current_directory,
            build=True,
        )

    def _run(self) -> None:
        Stop(
            self._arguments,
            self._original_directory,
            self._configuration,
            self._analysis_directory,
        ).run()
        # Force the incremental run to be blocking.
        # pyre-fixme[16]: `Namespace` has no attribute `nonblocking`.
        self._arguments.nonblocking = False
        # pyre-fixme[16]: `Namespace` has no attribute `no_start`.
        self._arguments.no_start = False
        Incremental(
            self._arguments,
            self._original_directory,
            self._configuration,
            self._analysis_directory,
        ).run()



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import argparse
import enum
import json
import logging
import os
import re
import resource
import signal
import subprocess
import threading
from abc import ABC, abstractmethod
from typing import Dict, Iterable, List, Optional, Set  # noqa

from .. import (
    find_local_root,
    find_log_directory,
    find_project_root,
    is_capable_terminal,
    json_rpc,
    log,
    readable_directory,
)
from ..analysis_directory import AnalysisDirectory, resolve_analysis_directory
from ..configuration import Configuration
from ..exceptions import EnvironmentException
from ..filesystem import remove_if_exists, translate_path
from ..socket_connection import SocketConnection, SocketException


TEXT = "text"  # type: str
JSON = "json"  # type: str


LOG = logging.getLogger(__name__)  # type: logging.Logger


class ClientException(Exception):
    pass


class State(enum.IntEnum):
    DEAD = 0
    RUNNING = 1


class ExitCode(enum.IntEnum):
    SUCCESS = 0
    FOUND_ERRORS = 1
    FAILURE = 2
    BUCK_ERROR = 3
    # If the process exited due to a signal, this will be the negative signal number.
    SIGSEGV = -signal.SIGSEGV


class IncrementalStyle(enum.Enum):
    SHALLOW = "shallow"
    TRANSITIVE = "transitive"
    FINE_GRAINED = "fine_grained"

    def __str__(self) -> str:
        return self.value


class ProfileOutput(enum.Enum):
    TRACE_EVENT: str = "trace_event"
    COLD_START_PHASES: str = "cold_start_phases"
    INCREMENTAL_UPDATES: str = "incremental_updates"

    def __str__(self) -> str:
        return self.value


class Result:
    def __init__(self, code: int, output: str) -> None:
        self.code = code
        self.output = output  # type: str

    def check(self) -> None:
        if self.code != ExitCode.SUCCESS:
            description = ":\n{}".format(self.output) if self.output else ""
            if self.code == ExitCode.SIGSEGV:
                description += (
                    "\nThis is a Pyre bug. Please re-run Pyre with --debug "
                    "and provide the output to the developers."
                )
            raise ClientException(
                "Client exited with error code {}{}".format(self.code, description)
            )


def typeshed_search_path(typeshed_root: str) -> List[str]:
    search_path = []
    typeshed_subdirectories = ["stdlib", "third_party"]
    for typeshed_subdirectory_name in typeshed_subdirectories:
        typeshed_subdirectory = os.path.join(typeshed_root, typeshed_subdirectory_name)
        if (
            not os.path.isdir(typeshed_subdirectory)
            or typeshed_subdirectory_name == "tests"
            or typeshed_subdirectory_name[0] == "."
        ):
            continue

        # Always prefer newer version over older version
        version_names = reversed(sorted(os.listdir(typeshed_subdirectory)))
        for version_name in version_names:
            # Anything under 2/ or 2.x is unusable for Pyre
            if version_name.startswith("2") and version_name != "2and3":
                continue
            search_path.append(os.path.join(typeshed_subdirectory, version_name))
    return search_path


def _convert_json_response_to_result(response: json_rpc.Response) -> Result:
    if response.error:
        error_code = ExitCode.FAILURE
    else:
        error_code = ExitCode.SUCCESS
    return Result(output=json.dumps(response.result), code=error_code)


def executable_file(file_path: str) -> str:
    if not os.path.isfile(file_path):
        raise EnvironmentException("%s is not a valid file" % file_path)
    if not os.access(file_path, os.X_OK):
        raise EnvironmentException("%s is not an executable file" % file_path)
    return file_path


class CommandParser(ABC):
    NAME = ""  # type: str
    _exit_code = ExitCode.SUCCESS  # type: ExitCode

    def __init__(self, arguments: argparse.Namespace, original_directory: str) -> None:
        self._arguments = arguments
        self._local_configuration: Optional[str] = arguments.local_configuration
        self._version: bool = arguments.version
        self._debug: bool = arguments.debug
        self._sequential: bool = arguments.sequential
        self._strict: bool = arguments.strict
        self._additional_checks: List[str] = arguments.additional_check
        self._show_error_traces: bool = arguments.show_error_traces
        self._output: str = arguments.output
        self._verbose: bool = arguments.verbose
        self._enable_profiling: bool = arguments.enable_profiling
        self._enable_memory_profiling: bool = arguments.enable_memory_profiling
        self._noninteractive: bool = arguments.noninteractive
        self._hide_parse_errors: bool = arguments.hide_parse_errors
        self._logging_sections: str = arguments.logging_sections
        self._log_identifier: str = arguments.log_identifier
        self._log_directory: str = arguments.log_directory
        self._logger: str = arguments.logger
        self._formatter: List[str] = arguments.formatter

        self._targets: List[str] = arguments.targets
        self._build: bool = arguments.build
        self._use_buck_builder: bool = arguments.use_buck_builder
        self._use_legacy_builder: bool = arguments.use_legacy_builder
        self._buck_builder_debug: bool = arguments.buck_builder_debug

        self._source_directories: List[str] = arguments.source_directories
        self._filter_directory: List[str] = arguments.filter_directory
        self._use_global_shared_analysis_directory: bool = arguments.use_global_shared_analysis_directory
        self._no_saved_state: bool = arguments.no_saved_state

        self._search_path: List[str] = arguments.search_path
        self._preserve_pythonpath: bool = arguments.preserve_pythonpath
        self._binary: str = arguments.binary
        self._buck_builder_binary: Optional[str] = arguments.buck_builder_binary
        self._buck_builder_target: Optional[str] = arguments.buck_builder_target
        self._exclude: List[str] = arguments.exclude
        self._typeshed: str = arguments.typeshed
        self._save_initial_state_to: Optional[str] = arguments.save_initial_state_to
        self._load_initial_state_from: Optional[str] = arguments.load_initial_state_from
        self._changed_files_path: Optional[str] = arguments.changed_files_path
        self._saved_state_project: Optional[str] = arguments.saved_state_project
        self._use_json_sockets: bool = arguments.use_json_sockets

        # Derived arguments
        self._capable_terminal: bool = is_capable_terminal()
        self._original_directory: str = original_directory
        self._current_directory: str = find_project_root(self._original_directory)
        self._local_configuration = self._local_configuration or find_local_root(
            self._original_directory
        )
        self._log_directory: str = find_log_directory(
            self._log_directory, self._current_directory, self._local_configuration
        )
        logger = self._logger
        if logger:
            self._logger = translate_path(self._original_directory, logger)
        if self._debug or not self._capable_terminal:
            self._noninteractive = True

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "-l", "--local-configuration", type=str, help="Use a local configuration"
        )

        parser.add_argument(
            "--version",
            action="store_true",
            help="Print the client and binary versions of Pyre.",
        )

        parser.add_argument("--debug", action="store_true", help=argparse.SUPPRESS)
        parser.add_argument("--sequential", action="store_true", help=argparse.SUPPRESS)
        parser.add_argument("--strict", action="store_true", help=argparse.SUPPRESS)
        parser.add_argument(
            "--additional-check", action="append", help=argparse.SUPPRESS
        )

        parser.add_argument(
            "--show-error-traces",
            action="store_true",
            help="Display errors trace information",
        )

        # Logging.
        parser.add_argument(
            "--output", choices=[TEXT, JSON], default=TEXT, help="How to format output"
        )
        parser.add_argument(
            "--verbose", action="store_true", help="Enable verbose logging"
        )
        parser.add_argument(
            "--enable-profiling", action="store_true", help=argparse.SUPPRESS
        )
        parser.add_argument(
            "--enable-memory-profiling", action="store_true", help=argparse.SUPPRESS
        )
        parser.add_argument(
            "-n",
            "--noninteractive",
            action="store_true",
            help="Disable interactive logging",
        )
        parser.add_argument(
            "--hide-parse-errors",
            action="store_true",
            help="Hide detailed information about parse errors",
        )
        parser.add_argument(
            "--logging-sections", help=argparse.SUPPRESS  # Enable sectional logging.
        )
        parser.add_argument(
            "--log-identifier",
            default="",
            help=argparse.SUPPRESS,  # Add given identifier to logged samples.
        )
        parser.add_argument(
            "--log-directory",
            help=argparse.SUPPRESS,  # Override default location for logs
        )
        parser.add_argument(
            "--logger", help=argparse.SUPPRESS  # Specify custom logging binary.
        )
        parser.add_argument("--formatter", help=argparse.SUPPRESS)

        # Link tree determination.
        buck_arguments = parser.add_argument_group("buck")
        buck_arguments.add_argument(
            "--target", action="append", dest="targets", help="The buck target to check"
        )
        buck_arguments.add_argument(
            "--build",
            action="store_true",
            help="Freshly build all the necessary artifacts.",
        )
        buck_arguments.add_argument(
            "--use-buck-builder",
            action="store_true",
            help="Use Pyre's experimental builder for Buck projects.",
        )
        buck_arguments.add_argument(
            "--use-legacy-builder",
            action="store_true",
            help="Use Pyre's legacy builder for Buck projects.",
        )
        buck_arguments.add_argument(
            "--buck-builder-debug", action="store_true", help=argparse.SUPPRESS
        )

        source_directories = parser.add_argument_group("source-directories")
        source_directories.add_argument(
            "--source-directory",
            action="append",
            dest="source_directories",
            help="The source directory to check",
            type=os.path.abspath,
        )
        source_directories.add_argument(
            "--filter-directory", help=argparse.SUPPRESS  # override filter directory
        )

        parser.add_argument(
            "--use-global-shared-analysis-directory",
            action="store_true",
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            "--no-saved-state",
            action="store_true",
            help="Don't attempt to load Pyre from a saved state.",
        )

        # Handling of search path
        parser.add_argument(
            "--search-path",
            action="append",
            default=[],
            type=readable_directory,
            help="Add an additional directory of modules and stubs to include"
            " in the type environment",
        )
        parser.add_argument(
            "--preserve-pythonpath",
            action="store_true",
            default=False,
            help="Preserve the value of the PYTHONPATH environment variable and "
            "inherit the current python environment's search path",
        )

        parser.add_argument(
            "--binary",
            default=None,
            type=executable_file,
            help="Location of the pyre binary",
        )

        parser.add_argument(
            "--buck-builder-binary",
            default=None,
            help="Location of the buck builder binary",
        )
        parser.add_argument(
            "--buck-builder-target", default=None, help=argparse.SUPPRESS
        )

        parser.add_argument(
            "--exclude",
            action="append",
            default=[],
            help="Exclude files and directories matching this regexp from parsing",
        )

        # Typeshed stubs location
        parser.add_argument(
            "--typeshed",
            default=None,
            type=readable_directory,
            help="Location of the typeshed stubs",
        )
        parser.add_argument(
            "--save-initial-state-to",
            default=None,
            help="Path to serialize pyre's initial state to.",
        )
        parser.add_argument(
            "--load-initial-state-from", default=None, type=str, help=argparse.SUPPRESS
        )
        parser.add_argument(
            "--changed-files-path", default=None, type=str, help=argparse.SUPPRESS
        )
        parser.add_argument(
            "--saved-state-project", default=None, type=str, help=argparse.SUPPRESS
        )
        parser.add_argument(
            "--features", default=None, type=str, help=argparse.SUPPRESS
        )
        # Temporary flag to help migrate to json sockets for incremental and query
        # commands.
        parser.add_argument(
            "--use-json-sockets",
            action="store_true",
            default=False,
            help=argparse.SUPPRESS,
        )

    @classmethod
    @abstractmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        pass

    @abstractmethod
    def _run(self) -> None:
        """ Abstract method expected to be overridden by subclasses. """
        pass

    def run(self) -> "CommandParser":
        self._run()
        return self

    def exit_code(self) -> int:
        return self._exit_code

    @property
    def current_directory(self) -> Optional[str]:
        return self._current_directory

    @property
    def local_configuration(self) -> Optional[str]:
        return self._local_configuration

    @property
    def log_directory(self) -> str:
        return self._log_directory

    @property
    def noninteractive(self) -> bool:
        return self._noninteractive


class Command(CommandParser):
    _buffer = []  # type: List[str]
    _call_client_terminated = False  # type: bool

    _local_root = ""  # type: str

    def __init__(
        self,
        arguments: argparse.Namespace,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super(Command, self).__init__(arguments, original_directory)
        local_configuration = self._local_configuration
        if local_configuration:
            self._local_root = (
                local_configuration
                if os.path.isdir(local_configuration)
                else os.path.dirname(local_configuration)
            )
        else:
            self._local_root = self._original_directory

        self._configuration: Configuration = configuration or self.generate_configuration()
        self._strict: bool = arguments.strict or self._configuration.strict
        self._logger: str = arguments.logger or (configuration and configuration.logger)
        self._ignore_all_errors_paths: Iterable[str] = (
            self._configuration.ignore_all_errors
        )
        self._number_of_workers: int = self._configuration.number_of_workers
        self._version_hash: str = self._configuration.version_hash
        self._formatter: Optional[str] = self._configuration.formatter
        self._taint_models_path: List[str] = [
            translate_path(self._original_directory, path)
            for path in self._configuration.taint_models_path
        ]

        self._analysis_directory: AnalysisDirectory = (
            analysis_directory or self.generate_analysis_directory()
        )
        self._features: Optional[str] = arguments.features

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        pass

    def generate_configuration(self) -> Configuration:
        return Configuration(
            local_configuration=self._local_configuration,
            search_path=self._search_path,
            binary=self._binary,
            typeshed=self._typeshed,
            preserve_pythonpath=self._preserve_pythonpath,
            excludes=self._exclude,
            logger=self._logger,
            formatter=self._formatter,
            log_directory=self._log_directory,
        )

    def generate_analysis_directory(self) -> AnalysisDirectory:
        configuration = self._configuration
        if not configuration:
            return AnalysisDirectory(".")
        else:
            return resolve_analysis_directory(
                self._arguments,
                configuration,
                self._original_directory,
                self._current_directory,
            )

    def run(self) -> "Command":
        configuration = self._configuration
        if configuration and configuration.disabled:
            LOG.log(log.SUCCESS, "Pyre will not run due to being explicitly disabled")
        else:
            self._run()
        return self

    def _run(self) -> None:
        pass

    def _flags(self) -> List[str]:
        flags = []
        if self._debug:
            flags.extend(["-debug"])
        if self._sequential:
            flags.extend(["-sequential"])
        if self._strict:
            flags.extend(["-strict"])
        if self._additional_checks:
            flags.append("-additional-checks")
            flags.append(",".join(self._additional_checks))
        if self._show_error_traces:
            flags.append("-show-error-traces")
        if self._verbose:
            flags.append("-verbose")
        if not self._hide_parse_errors:
            if self._logging_sections:
                self._logging_sections = self._logging_sections + ",parser"
            else:
                self._logging_sections = "parser"
        if not self._capable_terminal:
            # Disable progress reporting for non-capable terminals.
            # This helps in reducing clutter.
            if self._logging_sections:
                self._logging_sections = self._logging_sections + ",-progress"
            else:
                self._logging_sections = "-progress"
        if self._logging_sections:
            flags.extend(["-logging-sections", self._logging_sections])
        if self._enable_profiling:
            flags.extend(["-profiling-output", self.profiling_log_path()])
        if self._enable_memory_profiling:
            flags.extend(["-memory-profiling-output", self.profiling_log_path()])
        if self._enable_profiling or self._enable_memory_profiling:
            # Clear the profiling log first since in pyre binary it's append-only
            remove_if_exists(self.profiling_log_path())
        if self._current_directory:
            flags.extend(["-project-root", self._current_directory])
        if self._log_identifier:
            flags.extend(["-log-identifier", self._log_identifier])
        if self._logger:
            flags.extend(["-logger", self._logger])
        if self._log_directory:
            flags.extend(["-log-directory", self._log_directory])
        return flags

    # temporarily always return empty list to unblock client release
    def _feature_flags(self) -> List[str]:
        return []

    def _read_stdout(self, stdout: Iterable[bytes]) -> None:
        self._buffer = []
        for line in stdout:
            self._buffer.append(line.decode())

    def _read_stderr(self, stream: Iterable[bytes]) -> None:
        buffer = None
        log_pattern = re.compile(r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2} (\w+) (.*)")
        try:
            for line in stream:
                if self._call_client_terminated:
                    return
                line = line.decode().rstrip()
                match = log_pattern.match(line)
                if match:
                    if buffer:
                        buffer.flush()
                    buffer = log.Buffer(
                        section=match.groups()[0], data=[match.groups()[1]]
                    )
                elif buffer:
                    buffer.append(line)
            if buffer:
                buffer.flush()
        except Exception:
            pass

    def _call_client(self, command: str, capture_output: bool = True) -> Result:
        if not os.path.isdir(self._analysis_directory.get_root()):
            raise EnvironmentException(
                "`{}` is not a link tree.".format(self._analysis_directory.get_root())
            )

        client_command = [self._configuration.binary, command]
        client_command.extend(self._flags())
        client_command.append(self._analysis_directory.get_root())

        def limit_memory_usage() -> None:
            try:
                limit = 30 * 1024 * 1024 * 1024  # 30 GB
                resource.setrlimit(resource.RLIMIT_DATA, (limit, limit))
            except OSError:
                # Run the process with unlimited memory if the underlying syscall fails.
                pass

        LOG.debug("Running `%s`", " ".join(client_command))
        with subprocess.Popen(
            client_command,
            stdout=subprocess.PIPE if capture_output else None,
            stderr=subprocess.PIPE,
            preexec_fn=limit_memory_usage,
        ) as process:

            # Read stdout output
            if capture_output:
                stdout_reader = threading.Thread(
                    target=self._read_stdout, args=(process.stdout,)
                )
                stdout_reader.daemon = True
                stdout_reader.start()

            # Read the error output and print it.
            self._call_client_terminated = False
            stderr_reader = threading.Thread(
                target=self._read_stderr, args=(process.stderr,)
            )
            stderr_reader.daemon = True
            stderr_reader.start()

            # Wait for the process to finish and clean up.
            process.wait()
            # In the exceptional case, make sure that we print the error messages.
            if process.returncode != 0:
                stderr_reader.join()
            self._call_client_terminated = True
            if capture_output:
                # pyre-fixme: stdout_reader is not always declared!
                stdout_reader.join()

            output = ""
            if capture_output:
                output = "".join(self._buffer)
            if process.returncode != 0 and capture_output:
                output = "".join(self._buffer)

            return Result(code=process.returncode, output=output)

    def _relative_path(self, path: str) -> str:
        return os.path.relpath(path, self._original_directory)

    def _state(self) -> State:
        pid_path = os.path.join(self._log_directory, "server/server.pid")
        try:
            with open(pid_path) as file:
                pid = int(file.read())
                os.kill(pid, 0)  # throws if process is not running
            return State.RUNNING
        except Exception:
            return State.DEAD

    # will open a socket, send a request, read the response and close the socket.
    def _send_and_handle_socket_request(
        self, request: json_rpc.Request, version_hash: str
    ) -> None:
        try:
            with SocketConnection(self._log_directory) as socket_connection:
                socket_connection.perform_handshake(version_hash)
                socket_connection.send_request(request)
                response = json_rpc.read_response(socket_connection.input)
                result = _convert_json_response_to_result(response)
                result.check()
                self._socket_result_handler(result)
        except (
            SocketException,
            ResourceWarning,
            ClientException,
            json_rpc.JSONRPCException,
        ) as exception:
            LOG.error("Error while waiting for server: %s", str(exception))
            LOG.error("Run `pyre restart` in order to restart the server.")
            self._exit_code = ExitCode.FAILURE

    # Will be overwritten in subclasses to specialize how json socket
    # responses are handled.
    def _socket_result_handler(self, result: Result) -> None:
        log.stdout.write(result.output)

    def profiling_log_path(self) -> str:
        return os.path.join(self._log_directory, "profiling.log")

    @property
    def analysis_directory(self) -> AnalysisDirectory:
        return self._analysis_directory



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import argparse
import json
import logging
import os
import shutil
import subprocess
import sys
from logging import Logger
from typing import Any, Dict

from .. import (
    BINARY_NAME,
    CONFIGURATION_FILE,
    find_taint_models_directory,
    find_typeshed,
    log,
)
from ..exceptions import EnvironmentException
from .command import CommandParser


LOG: Logger = logging.getLogger(__name__)


class Initialize(CommandParser):
    NAME = "initialize"

    def __init__(self, arguments, original_directory: str) -> None:
        super(Initialize, self).__init__(arguments, original_directory)
        self._local: bool = arguments.local

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        initialize = parser.add_parser(cls.NAME, aliases=["init"])
        initialize.set_defaults(command=cls)
        initialize.add_argument(
            "--local",
            action="store_true",
            help="Initializes a local configuration in a project subdirectory.",
        )

    def _get_configuration(self) -> Dict[str, Any]:
        configuration = {}  # type: Dict[str, Any]

        watchman_configuration_path = os.path.abspath(".watchmanconfig")
        if shutil.which("watchman") is not None and log.get_yes_no_input(
            "Also initialize a watchman configuration?"
        ):
            try:
                with open(watchman_configuration_path, "w+") as configuration_file:
                    configuration_file.write("{}\n")
                subprocess.check_call(["watchman", "watch-project", "."])
            except (IsADirectoryError, subprocess.CalledProcessError):
                LOG.warning("Unable to initialize watchman for the current directory.")

        binary_path = shutil.which(BINARY_NAME)
        if binary_path is None:
            binary_path = shutil.which(
                os.path.join(os.path.dirname(sys.argv[0]), BINARY_NAME)
            )
        if binary_path is None:
            binary_path = os.path.abspath(
                log.get_input(
                    "No `{}` found, enter the path manually: ".format(BINARY_NAME)
                )
            )
            if not os.path.isfile(binary_path):
                raise EnvironmentException(
                    "Unable to locate binary at `{}`.".format(binary_path)
                )
        else:
            LOG.info("Binary found at `{}`".format(binary_path))
        configuration["binary"] = binary_path

        typeshed = find_typeshed()
        if typeshed is None:
            typeshed = os.path.abspath(
                log.get_input("Unable to locate typeshed, please enter its root: ")
            )
            if not os.path.isdir(typeshed):
                raise EnvironmentException(
                    "No typeshed directory found at `{}`.".format(typeshed)
                )
        configuration["typeshed"] = typeshed

        taint_models_path = find_taint_models_directory()
        if taint_models_path is not None:
            configuration["taint_models_path"] = taint_models_path

        analysis_directory = log.get_optional_input(
            "Which directory should pyre be initialized in?", "."
        )

        configuration["source_directories"] = [analysis_directory]
        return configuration

    def _get_local_configuration(self) -> Dict[str, Any]:
        configuration = {}  # type: Dict[str, Any]
        targets = log.get_input(
            "Which buck target(s) should pyre analyze? (//target:a,//target/b/...)\n"
        )
        configuration["targets"] = [target.strip() for target in targets.split(",")]
        continuous = log.get_yes_no_input(
            "Would you like to enable Pyre's continuous integration for your changes?"
        )
        configuration["continuous"] = continuous
        if continuous:
            push_blocking = log.get_yes_no_input(
                "Would you like the continuous integration to be push blocking?"
            )
            configuration["push_blocking"] = push_blocking
            if push_blocking:
                # Push blocking implies continuous, it's confusing to have both.
                del configuration["continuous"]
                configuration["differential"] = log.get_yes_no_input(
                    "Should pyre only be push-blocking on newly introduced errors?"
                )
        return configuration

    def _run(self) -> None:
        configuration_path = os.path.join(self._original_directory, CONFIGURATION_FILE)
        if os.path.isfile(configuration_path):
            if self._local:
                error = "Local configurations must be created in subdirectories of `{}`"
                "as it already contains a `.pyre_configuration`.".format(
                    self._original_directory
                )
            else:
                error = "A pyre configuration already exists at `{}`.".format(
                    configuration_path
                )
            raise EnvironmentException(error)
        if os.path.isfile(configuration_path + ".local"):
            raise EnvironmentException(
                "A local pyre configuration already exists at `{}`.".format(
                    configuration_path + ".local"
                )
            )
        if self._local:
            configuration_path = configuration_path + ".local"
            configuration = self._get_local_configuration()
        else:
            configuration = self._get_configuration()

        with open(configuration_path, "w+") as configuration_file:
            json.dump(configuration, configuration_file, sort_keys=True, indent=2)
            configuration_file.write("\n")
        LOG.info(
            "Successfully initialized pyre! "
            + "You can view the configuration at `{}`.".format(configuration_path)
        )



# Copyright 2004-present Facebook.  All rights reserved.

# pyre-unsafe

import functools
import glob
import json
import logging
import os
import subprocess
import sys
import tempfile
import threading
from collections import namedtuple
from json.decoder import JSONDecodeError
from logging import Logger
from typing import Dict, Iterable, List, Optional, Set, Tuple, Union, cast  # noqa

from .filesystem import BuckBuilder, find_root


LOG: Logger = logging.getLogger(__name__)
BuckOut = namedtuple("BuckOut", "source_directories targets_not_found")


class BuckException(Exception):
    pass


class FastBuckBuilder(BuckBuilder):
    def __init__(
        self,
        buck_root: str,
        output_directory: Optional[str] = None,
        buck_builder_binary: Optional[str] = None,
        buck_builder_target: Optional[str] = None,
        debug_mode: bool = False,
    ) -> None:
        self._buck_root = buck_root
        self._output_directory = output_directory or tempfile.mkdtemp(
            prefix="pyre_tmp_"
        )
        self._buck_builder_binary = buck_builder_binary
        self._buck_builder_target = buck_builder_target
        self._debug_mode = debug_mode
        self.conflicting_files = []
        self.unsupported_files = []

    def _get_builder_executable(self) -> str:
        builder_binary = self._buck_builder_binary
        if not self._debug_mode:
            if builder_binary is None:
                raise BuckException(
                    "--buck-builder-binary must be provided "
                    "if --buck-builder-debug is not enabled."
                )
            return builder_binary
        target = self._buck_builder_target
        if target is None:
            raise BuckException(
                "--buck-builder-target must be provided "
                "if --buck-builder-debug is enabled."
            )
        binary_relative_path = (
            subprocess.check_output(
                [
                    "buck",
                    "build",
                    "--show-output",
                    "//tools/pyre/facebook/fb_buck_project_builder",
                ],
                stderr=subprocess.DEVNULL,
            )
            .decode()
            .strip()
            .split(" ")[1]
        )
        return os.path.join(self._buck_root, binary_relative_path)

    def build(self, targets: Iterable[str]) -> List[str]:
        command = [
            self._get_builder_executable(),
            "-J-Djava.net.preferIPv6Addresses=true",
            "-J-Djava.net.preferIPv6Stack=true",
            "--buck_root",
            self._buck_root,
            "--output_directory",
            self._output_directory,
        ] + list(targets)
        if self._debug_mode:
            command.append("--debug")
        LOG.info("Buck builder command: `{}`".format(" ".join(command)))
        with subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        ) as buck_builder_process:
            # Java's logging conflicts with Python's logging, we capture the
            # logs and re-log them with python's logger.
            log_processor = threading.Thread(
                target=self._read_stderr, args=(buck_builder_process.stderr,)
            )
            log_processor.daemon = True
            log_processor.start()
            return_code = buck_builder_process.wait()
            # Wait until all stderr have been printed.
            log_processor.join()
            if return_code == 0:
                LOG.info("Finished building targets.")
                if self._debug_mode:
                    debug_output = json.loads(
                        "".join([line.decode() for line in buck_builder_process.stdout])
                    )
                    self.conflicting_files += debug_output["conflictingFiles"]
                    self.unsupported_files += debug_output["unsupportedFiles"]
                return [self._output_directory]
            else:
                raise BuckException(
                    "Could not build targets. Check the paths or run `buck clean`."
                )

    def _read_stderr(self, stream: Iterable[bytes]) -> None:
        for line in stream:
            line = line.decode().rstrip()
            if line.startswith("INFO: "):
                LOG.info(line[6:])
            elif line.startswith("WARNING: "):
                LOG.warning(line[9:])
            elif line.startswith("ERROR: "):
                LOG.error(line[7:])
            elif line.startswith("[WARNING:"):
                # Filter away thrift warnings.
                pass
            else:
                LOG.error(line)


class SimpleBuckBuilder(BuckBuilder):
    def __init__(self, build: bool = True) -> None:
        self._build = build

    def build(self, targets: Iterable[str]) -> Iterable[str]:
        """
            Shell out to buck to build the targets, then yield the paths to the
            link trees.
        """
        return generate_source_directories(targets, build=self._build)


def presumed_target_root(target):
    root_index = target.find("//")
    if root_index != -1:
        target = target[root_index + 2 :]
    target = target.replace("/...", "")
    target = target.split(":")[0]
    return target


# Expects the targets to be already normalized.
def _find_built_source_directories(
    targets_to_destinations: Iterable[Tuple[str, str]]
) -> BuckOut:
    targets_not_found = []
    source_directories = []
    buck_root = find_buck_root(os.getcwd())
    if buck_root is None:
        raise Exception("No .buckconfig found in ancestors of the current directory.")

    directories = set()
    for target, destination in targets_to_destinations:
        directories.add((target, os.path.dirname(destination)))

    for target, directory in directories:
        target_name = target.split(":")[1]
        discovered_source_directories = glob.glob(
            os.path.join(buck_root, directory, "{}#*link-tree".format(target_name))
        )
        if len(discovered_source_directories) == 0:
            targets_not_found.append(target)
        source_directories.extend(
            [
                tree
                for tree in discovered_source_directories
                if not tree.endswith(
                    (
                        "-vs_debugger#link-tree",
                        "-interp#link-tree",
                        "-ipython#link-tree",
                    )
                )
            ]
        )
    return BuckOut(set(source_directories), set(targets_not_found))


def _normalize(targets: List[str]) -> List[Tuple[str, str]]:
    LOG.info(
        "Normalizing target%s `%s`",
        "s:" if len(targets) > 1 else "",
        "`, `".join(targets),
    )
    try:
        command = (
            ["buck", "targets", "--show-output"]
            + targets
            + ["--type", "python_binary", "python_test"]
        )
        targets_to_destinations = (
            subprocess.check_output(command, stderr=subprocess.PIPE, timeout=600)
            .decode()
            .strip()
            .split("\n")
        )  # type: List[str]
        targets_to_destinations = list(filter(bool, targets_to_destinations))
        # The output is of the form //target //corresponding.par
        result = []
        for target in targets_to_destinations:
            pair = target.split(" ")
            if len(pair) != 2:
                pass
            else:
                result.append((pair[0], pair[1]))
        if not result:
            LOG.warning(
                "Provided targets do not contain any binary or unittest targets."
            )
            return []
        else:
            LOG.info(
                "Found %d buck target%s.", len(result), "s" if len(result) > 1 else ""
            )
        return result
    except subprocess.TimeoutExpired as error:
        LOG.error("Buck output so far: %s", error.stderr.decode().strip())
        raise BuckException(
            "Seems like `{}` is hanging.\n   "
            "Try running `buck clean` before trying again.".format(
                # pyre-fixme: command not always defined
                " ".join(command[:-1])
            )
        )
    except subprocess.CalledProcessError as error:
        LOG.error("Buck returned error: %s" % error.stderr.decode().strip())
        raise BuckException(
            "Could not normalize targets. Check the paths or run `buck clean`."
        )


def _build_targets(targets: List[str], original_targets: List[str]) -> None:
    LOG.info(
        "Building target%s `%s`",
        "s:" if len(original_targets) > 1 else "",
        "`, `".join(original_targets),
    )
    command = ["buck", "build"] + targets
    try:
        subprocess.check_output(command, stderr=subprocess.PIPE)
        LOG.warning("Finished building targets.")
    except subprocess.CalledProcessError as error:
        # The output can be overwhelming, hence print only the last 20 lines.
        lines = error.stderr.decode().splitlines()
        LOG.error("Buck returned error: %s" % "\n".join(lines[-20:]))
        raise BuckException(
            "Could not build targets. Check the paths or run `buck clean`."
        )


def _map_normalized_targets_to_original(
    unbuilt_targets: Iterable[str], original_targets: Iterable[str]
) -> List[str]:
    mapped_targets = set()
    for target in unbuilt_targets:
        # Each original target is either a `/...` glob or a proper target.
        # If it's a glob, we're looking for the glob to be a prefix of the unbuilt
        # target. Otherwise, we care about exact matches.
        name = None
        for original in original_targets:
            if original.endswith("/..."):
                if target.startswith(original[:-4]):
                    name = original
            else:
                if target == original:
                    name = original
        # No original target matched, fallback to normalized.
        if name is None:
            name = target
        mapped_targets.add(name)
    return list(mapped_targets)


@functools.lru_cache()
def find_buck_root(path: str) -> Optional[str]:
    return find_root(path, ".buckconfig")


def query_buck_relative_paths(
    project_paths: Iterable[str], targets: Iterable[str]
) -> Dict[str, str]:
    """Return a mapping from each absolute project path to its relative location
    in the buck output directory.
    This queries buck and only returns paths that are covered by `targets`."""
    buck_root = find_buck_root(os.getcwd())
    if buck_root is None:
        LOG.error(
            "Buck root couldn't be found. Returning empty analysis directory mapping."
        )
        return {}
    target_string = " ".join(targets)
    command = [
        "buck",
        "query",
        "--json",
        "--output-attribute",
        ".*",
        # This will get only those owner targets that are beneath our targets or
        # the dependencies of our targets.
        f"owner(%s) ^ deps(set({target_string}))",
        *project_paths,
    ]
    LOG.info(f"Running command: {command}")
    try:
        owner_output = json.loads(
            subprocess.check_output(command, timeout=30, stderr=subprocess.DEVNULL)
            .decode()
            .strip()
        )
    except (
        subprocess.TimeoutExpired,
        subprocess.CalledProcessError,
        JSONDecodeError,
    ) as error:
        raise BuckException("Querying buck for relative paths failed: {}".format(error))

    results = {}
    for project_path in project_paths:
        for target_data in owner_output.values():
            prefix = os.path.join(buck_root, target_data["buck.base_path"]) + os.sep
            suffix = project_path[len(prefix) :]
            if not project_path.startswith(prefix) or suffix not in target_data["srcs"]:
                continue

            if "buck.base_module" in target_data:
                base_path = os.path.join(*target_data["buck.base_module"].split("."))
            else:
                base_path = target_data["buck.base_path"]
            results[project_path] = os.path.join(base_path, target_data["srcs"][suffix])
            # Break after the first one because there might be multiple matches.
            break
    return results


def generate_source_directories(
    original_targets: Iterable[str], build: bool
) -> Set[str]:
    original_targets = list(original_targets)
    targets_to_destinations = _normalize(original_targets)
    targets = [pair[0] for pair in targets_to_destinations]
    if build:
        _build_targets(targets, original_targets)
    buck_out = _find_built_source_directories(targets_to_destinations)
    source_directories = buck_out.source_directories

    if buck_out.targets_not_found:
        if not build:
            # Build all targets to ensure buck doesn't remove some link trees as we go.
            _build_targets(targets, original_targets)
            buck_out = _find_built_source_directories(targets_to_destinations)
            source_directories = buck_out.source_directories

    if buck_out.targets_not_found:
        message_targets = _map_normalized_targets_to_original(
            buck_out.targets_not_found, original_targets
        )

        raise BuckException(
            "Could not find link trees for:\n    `{}`.\n   "
            "See `{} --help` for more information.".format(
                "    \n".join(message_targets), sys.argv[0]
            )
        )

    return source_directories



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import errno
import fcntl
import functools
import logging
import os
import shutil
import subprocess
from contextlib import contextmanager
from typing import Dict, Generator, Iterable, List, Optional, Set

from .exceptions import EnvironmentException


LOG: logging.Logger = logging.getLogger(__name__)


class BuckBuilder:
    def build(self, targets: Iterable[str]) -> Iterable[str]:
        """
            Build the given targets, and return a list of output directories
            containing the target output.
        """
        raise NotImplementedError


def translate_path(root: str, path: str) -> str:
    if os.path.isabs(path):
        return path

    translated = os.path.join(root, path)
    if os.path.exists(translated):
        return os.path.realpath(translated)

    return path


def translate_paths(paths: Set[str], original_directory: str) -> Set[str]:
    current_directory = os.getcwd()
    if not original_directory.startswith(current_directory):
        return paths
    translation = os.path.relpath(original_directory, current_directory)
    if not translation:
        return paths
    return {translate_path(translation, path) for path in paths}


def find_root(original_directory: str, target_file: str) -> Optional[str]:
    current_directory = os.path.abspath(original_directory)
    while current_directory != "/":
        absolute = os.path.join(current_directory, target_file)
        if os.path.isfile(absolute):
            return current_directory
        current_directory = os.path.dirname(current_directory)
    return None


def exists(path: str) -> str:
    if not os.path.isfile(path):
        raise ValueError("%s is not a valid file" % path)
    return path


def is_parent(parent: str, child: str) -> bool:
    return child.startswith(parent.rstrip(os.sep) + os.sep)


def find_paths_with_extensions(root: str, extensions: Iterable[str]) -> List[str]:
    root = os.path.abspath(root)  # Return absolute paths.
    extension_filter = []
    for extension in extensions:
        if len(extension_filter) > 0:
            extension_filter.append("-or")
        extension_filter.extend(["-name", "*.{}".format(extension)])

    output = (
        subprocess.check_output(
            [
                "find",
                root,
                # All files ending with the given extensions ...
                "(",
                *extension_filter,
                ")",
                # ... and that are either regular files ...
                "(",
                "-type",
                "f",
                "-or",
                # ... or symlinks.
                "-type",
                "l",
                ")",
                # Print all such files.
                "-print",
            ],
            stderr=subprocess.DEVNULL,
        )
        .decode("utf-8")
        .strip()
    )
    return output.split("\n") if output else []


def find_python_paths(root: str) -> List[str]:
    try:
        return find_paths_with_extensions(root, ["py", "pyi"])
    except subprocess.CalledProcessError:
        raise EnvironmentException(
            "Pyre was unable to locate an analysis directory. "
            "Ensure that your project is built and re-run pyre."
        )


def is_empty(path: str) -> bool:
    try:
        return os.stat(path).st_size == 0
    except FileNotFoundError:
        return False


def remove_if_exists(path: str) -> None:
    try:
        os.remove(path)
    except OSError:
        pass  # Not a file.
    try:
        shutil.rmtree(path)
    except OSError:
        pass  # Not a directory.


def _compute_symbolic_link_mapping(
    directory: str, extensions: Iterable[str]
) -> Dict[str, str]:
    """
        Given a shared analysis directory, produce a mapping from actual source files
        to files contained within this directory. Only includes files which have
        one of the provided extensions.

        Watchman watches actual source files, so when a change is detected to a
        file, this mapping can be used to identify what file changed from Pyre's
        perspective.
    """
    symbolic_links = {}
    try:
        for symbolic_link in find_paths_with_extensions(directory, extensions):
            symbolic_links[os.path.realpath(symbolic_link)] = symbolic_link
    except subprocess.CalledProcessError as error:
        LOG.warning(
            "Exception encountered trying to find source files "
            "in the analysis directory: `%s`",
            error,
        )
        LOG.warning("Starting with an empty set of tracked files.")
    return symbolic_links


def _delete_symbolic_link(link_path: str) -> None:
    os.unlink(link_path)


def add_symbolic_link(link_path: str, actual_path: str) -> None:
    directory = os.path.dirname(link_path)
    try:
        os.makedirs(directory)
    except OSError:
        pass
    try:
        os.symlink(actual_path, link_path)
    except OSError as error:
        if error.errno == errno.EEXIST:
            os.unlink(link_path)
            os.symlink(actual_path, link_path)
        else:
            LOG.error(str(error))


@contextmanager
def acquire_lock(path: str, blocking: bool) -> Generator[Optional[int], None, None]:
    """Raises an OSError if the lock can't be acquired"""
    try:
        with open(path, "w+") as lockfile:
            if not blocking:
                lock_command = fcntl.LOCK_EX | fcntl.LOCK_NB
            else:
                lock_command = fcntl.LOCK_EX

            fcntl.lockf(lockfile.fileno(), lock_command)
            yield lockfile.fileno()
            fcntl.lockf(lockfile.fileno(), fcntl.LOCK_UN)

    except FileNotFoundError:
        yield


class Filesystem:
    def list(
        self, root: str, patterns: List[str], exclude: Optional[List[str]] = None
    ) -> List[str]:
        """
            Return the list of files that match any of the patterns within root.
            If exclude is provided, files that match an exclude pattern are omitted.

            Note: The `find` command does not understand globs properly.
                e.g. 'a/*.py' will match 'a/b/c.py'
            For this reason, avoid calling this method with glob patterns.
        """

        command = ["find", "."]
        command += self._match_any(patterns)
        if exclude:
            command += ["-and", "!"]
            command += self._match_any(exclude)
        return (
            subprocess.run(command, stdout=subprocess.PIPE, cwd=root)
            .stdout.decode("utf-8")
            .split()
        )

    def _match_any(self, patterns: List[str]) -> List[str]:
        expression = []
        for pattern in patterns:
            if expression:
                expression.append("-or")
            expression.extend(["-path", "./{}".format(pattern)])
        return ["(", *expression, ")"]


class MercurialBackedFilesystem(Filesystem):
    def list(
        self, root: str, patterns: List[str], exclude: Optional[List[str]] = None
    ) -> List[str]:
        try:
            command = ["hg", "files"]
            for pattern in patterns:
                command += ["--include", pattern]
            if exclude:
                for pattern in exclude:
                    command += ["--exclude", pattern]
            return (
                subprocess.run(
                    command, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, cwd=root
                )
                .stdout.decode("utf-8")
                .split()
            )
        except FileNotFoundError:
            raise EnvironmentException("hg executable not found.")


@functools.lru_cache(1)
def get_filesystem() -> Filesystem:
    try:
        subprocess.check_output(["hg", "status"], stderr=subprocess.DEVNULL)
        return MercurialBackedFilesystem()
    except (subprocess.CalledProcessError, FileNotFoundError):
        return Filesystem()



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
from typing import List

from . import Check


class Deobfuscate(Check):
    NAME = "deobfuscate"  # type: str

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        deobfuscate = parser.add_parser(cls.NAME)
        deobfuscate.set_defaults(command=cls)

    def _flags(self) -> List[str]:
        flags = super()._flags()
        flags.extend(["-additional-check", "deobfuscation"])
        return flags



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import functools
import logging
import subprocess
from pathlib import Path
from typing import List

from mypy_extensions import TypedDict

from .. import LOG_DIRECTORY, find_project_root
from .command import Command


LOG = logging.getLogger(__name__)  # type: logging.Logger


ROOT_PLACEHOLDER_NAME = "<root>"
PID_MAXIMUM_WIDTH = 10


class ServerDetails(TypedDict):
    local_root: str
    pid: int
    path: Path


class Servers(Command):
    NAME = "servers"

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        servers_parser = parser.add_parser(
            cls.NAME,
            epilog="""
            Command to manipulate multiple Pyre servers.
            """,
        )
        servers_parser.set_defaults(command=cls, noninteractive=True)
        servers_parser.add_argument("list", help="List running servers.")

    @staticmethod
    @functools.lru_cache()
    def _dot_pyre_root() -> Path:
        return Path(find_project_root(str(Path.cwd())), LOG_DIRECTORY)

    @staticmethod
    def _print_server_details(all_server_details: List[ServerDetails]) -> None:
        print("Pyre servers: <pid> <server-path>")
        for details in all_server_details:
            print(
                "{:<{column_one_width}} {}".format(
                    details["pid"],
                    details["local_root"],
                    column_one_width=PID_MAXIMUM_WIDTH,
                )
            )

    @staticmethod
    def _find_servers() -> List[Path]:
        find_command = [
            "find",
            str(Servers._dot_pyre_root()),
            "-type",
            "f",
            "-name",
            "server.pid",
        ]
        LOG.info("Running command: `{}`".format(" ".join(find_command)))
        try:
            server_paths = subprocess.check_output(find_command).decode().splitlines()
        except subprocess.CalledProcessError as exception:
            LOG.error(f"`find` failed with exception `{exception}`")
            raise exception
        return [Path(path) for path in server_paths]

    @staticmethod
    def _fetch_server_details(server_path: Path, dot_pyre_root: Path) -> ServerDetails:
        local_root = str(server_path.relative_to(dot_pyre_root).parent.parent)
        return {
            "pid": int(server_path.read_text()),
            "path": server_path,
            "local_root": ROOT_PLACEHOLDER_NAME if local_root == "." else local_root,
        }

    def _run(self) -> None:
        server_paths = self._find_servers()
        all_server_details = sorted(
            (
                self._fetch_server_details(server_path, Servers._dot_pyre_root())
                for server_path in server_paths
            ),
            key=lambda details: details["local_root"],
        )
        self._print_server_details(all_server_details)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe
import argparse
from typing import List, Optional

from .. import get_binary_version
from ..analysis_directory import AnalysisDirectory
from ..configuration import Configuration
from ..version import __version__
from .command import Command


class Rage(Command):
    NAME = "rage"

    def __init__(
        self,
        arguments,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super(Rage, self).__init__(
            arguments, original_directory, configuration, analysis_directory
        )

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        rage = parser.add_parser(
            cls.NAME,
            epilog="""
            Collects troubleshooting diagnostics for Pyre, and writes this
            information to the terminal.
            """,
        )
        rage.set_defaults(command=cls)

    def _flags(self) -> List[str]:
        log_directory = self._log_directory
        if log_directory:
            return ["-log-directory", log_directory]
        else:
            return []

    def _run(self) -> None:
        # Do not use logging. Logging goes to stderr.
        print("Client version:", __version__, flush=True)
        print("Binary path:", self._configuration.binary, flush=True)
        print(
            "Configured binary version:",
            get_binary_version(self._configuration),
            flush=True,
        )
        self._call_client(command=self.NAME, capture_output=False).check()



from typing import Any, Dict, Iterable, List, NamedTuple, Optional

from .connection_api import PyreConnection


class DefineParameter(NamedTuple):
    name: str
    annotation: str


class Define(NamedTuple):
    name: str
    parameters: List[DefineParameter]
    return_annotation: str


class Position(NamedTuple):
    line: int
    column: int


class Location(NamedTuple):
    path: str
    start: Position
    stop: Position


class CallGraphTarget(NamedTuple):
    target: str
    # We might want to turn this into an enum in the future.
    kind: str
    locations: List[Location]


def defines(pyre_connection: PyreConnection, modules: Iterable[str]) -> List[Define]:
    query = "defines({})".format(",".join(modules))
    result = pyre_connection.query_server(query)
    if result is None or "response" not in result:
        return []
    return [
        Define(
            name=element["name"],
            parameters=[
                DefineParameter(
                    name=parameter["name"], annotation=parameter["annotation"]
                )
                for parameter in element["parameters"]
            ],
            return_annotation=element["return_annotation"],
        )
        for element in result["response"]
    ]


def get_class_hierarchy(
    pyre_connection: PyreConnection
) -> Optional[Dict[str, List[str]]]:
    result = pyre_connection.query_server("dump_class_hierarchy()")
    if result is None or "response" not in result:
        return None
    hierarchy = {
        key: edges
        for annotation_and_edges in result["response"]
        for key, edges in annotation_and_edges.items()
    }
    return hierarchy


def get_call_graph(
    pyre_connection: PyreConnection
) -> Optional[Dict[str, List[CallGraphTarget]]]:
    result = pyre_connection.query_server("dump_call_graph()")
    if result is None or "response" not in result:
        return None
    call_graph = {}
    for function, calls in result["response"].items():
        call_graph[function] = [
            CallGraphTarget(
                target=call["target"],
                kind=call["kind"],
                locations=[_parse_location(location) for location in call["locations"]],
            )
            for call in calls
        ]
    return call_graph


def _parse_location(location_json: Dict[str, Any]) -> Location:
    return Location(
        path=location_json["path"],
        start=_parse_position(location_json["start"]),
        stop=_parse_position(location_json["stop"]),
    )


def _parse_position(position_json: Dict[str, Any]) -> Position:
    return Position(line=position_json["line"], column=position_json["column"])



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import logging
import os
import shutil
import signal
import subprocess
from typing import Optional

import psutil

from .. import BINARY_NAME, CLIENT_NAME
from ..analysis_directory import AnalysisDirectory
from ..configuration import Configuration
from ..project_files_monitor import ProjectFilesMonitor
from .command import Command


LOG = logging.getLogger(__name__)  # type: logging.Logger


class Kill(Command):
    NAME = "kill"

    def __init__(
        self,
        arguments: argparse.Namespace,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super(Kill, self).__init__(
            arguments, original_directory, configuration, analysis_directory
        )
        self._with_fire: bool = arguments.with_fire

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        kill = parser.add_parser(cls.NAME)
        kill.set_defaults(command=cls)
        kill.add_argument(
            "--with-fire", action="store_true", help="Adds emphasis to the command."
        )

    def generate_analysis_directory(self) -> AnalysisDirectory:
        return AnalysisDirectory(".")

    @staticmethod
    def _delete_linked_path(link_path: str) -> None:
        try:
            actual_path = os.readlink(link_path)
            os.remove(actual_path)
        except OSError:
            pass
        try:
            os.unlink(link_path)
        except OSError:
            pass

    def _delete_caches(self) -> None:
        # If a resource cache exists, delete it to remove corrupted artifacts.
        try:
            shutil.rmtree(os.path.join(self._log_directory, "resource_cache"))
        except OSError:
            pass
        # If a buck builder cache exists, also remove it.
        try:
            shutil.rmtree("/tmp/pyre/buck_builder_cache")
        except OSError:
            pass

    def _kill_client_processes(self) -> None:
        for process in psutil.process_iter(attrs=["name"]):
            if process.info["name"] != CLIENT_NAME:
                continue
            # We need to be careful about how we kill the client here, as otherwise we
            # might cause a race where we attempt to kill the `pyre kill` command.
            pid_to_kill = process.pid
            if pid_to_kill == os.getpgid(os.getpid()):
                continue
            try:
                LOG.info(
                    "Killing process {} with pid {}.".format(
                        process.info["name"], pid_to_kill
                    )
                )
                os.kill(pid_to_kill, signal.SIGKILL)
            except ProcessLookupError:
                continue
        ProjectFilesMonitor.stop_project_monitor(self._configuration)

    @staticmethod
    def _kill_binary_processes() -> None:
        # Kills all processes that have the same binary as the one specified
        # in the configuration.
        binary_name = _get_process_name("PYRE_BINARY", BINARY_NAME)
        subprocess.run(["pkill", binary_name])

    def _run(self) -> None:
        self._kill_binary_processes()

        server_root = os.path.join(self._log_directory, "server")
        self._delete_linked_path(os.path.join(server_root, "server.sock"))
        self._delete_linked_path(os.path.join(server_root, "json_server.sock"))

        if self._arguments.with_fire is True:
            self._delete_caches()
        self._kill_client_processes()


def _get_process_name(environment_variable_name: str, default: str) -> str:
    overridden = os.getenv(environment_variable_name)
    if overridden is not None:
        return os.path.basename(overridden)
    else:
        return default



import glob
import os
import sys
from typing import List, Tuple

from setuptools import find_packages, setup


if sys.version_info < (3, 5):
    sys.exit("Error: pyre-check only runs on Python 3.5 and above.")


def get_data_files(base, directory, extension_glob) -> Tuple[str, List[str]]:
    # We need to relativize data_files, see https://github.com/pypa/wheel/issues/92.
    relative_directory = os.path.relpath(os.path.abspath(directory), base)
    return (
        relative_directory,
        glob.glob(os.path.join(relative_directory, extension_glob)),
    )


def find_typeshed_files():
    base = os.getcwd()
    typeshed_root = os.path.join(base, "typeshed")
    if not os.path.isdir(typeshed_root):
        return []
    result = []
    for absolute_directory, _, _ in os.walk(typeshed_root):
        relative_directory, files = get_data_files(
            base=base, directory=absolute_directory, extension_glob="*.pyi"
        )
        if not files:
            continue
        target = os.path.join("lib", "pyre_check", relative_directory)
        result.append((target, files))
    return result


def find_taint_stubs():
    base = os.getcwd()
    _, taint_stubs = get_data_files(
        base=base, directory=os.path.join(base, "taint"), extension_glob="*"
    )
    _, third_party_taint_stubs = get_data_files(
        base=base, directory=os.path.join(base, "third_party_taint"), extension_glob="*"
    )
    taint_stubs += third_party_taint_stubs
    if not taint_stubs:
        return []

    return [(os.path.join("lib", "pyre_check", "taint"), taint_stubs)]


with open("README.md") as f:
    long_description = f.read()


PYRE_EMAIL = "pyre@fb.com"
AUTHOR = "Facebook"

setup(
    name="pyre-check",
    version=os.environ["PACKAGE_VERSION"],
    description="A performant type checker for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://pyre-check.org/",
    download_url="https://github.com/facebook/pyre-check",
    author=AUTHOR,
    author_email=PYRE_EMAIL,
    maintainer=AUTHOR,
    maintainer_email=PYRE_EMAIL,
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development",
    ],
    keywords="typechecker development",
    packages=find_packages(exclude=["tests", "pyre-check"]),
    data_files=[("bin", ["bin/pyre.bin"])] + find_typeshed_files() + find_taint_stubs(),
    python_requires=">=3.5",
    install_requires=["pywatchman", "psutil", "libcst", "pyre_extensions"],
    entry_points={
        "console_scripts": [
            "pyre = pyre_check.client.pyre:main",
            "pyre-upgrade = pyre_check.tools.upgrade.upgrade:main",
        ]
    },
)



# Copyright (c) 2019-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import functools
import logging
import os
from typing import Any, BinaryIO, Dict, Iterable, List, Optional, Set  # noqa

from . import json_rpc
from .analysis_directory import AnalysisDirectory
from .configuration import Configuration
from .filesystem import find_root
from .socket_connection import SocketConnection
from .watchman_subscriber import Subscription, WatchmanSubscriber


LOG = logging.getLogger(__name__)  # type: logging.Logger


class MonitorException(Exception):
    pass


class ProjectFilesMonitor(WatchmanSubscriber):
    """
        Logs from this monitor are found in
        .pyre/<local root>/file_monitor/file_monitor.log
        One file monitor is spawned per pyre server. When a server is stopped,
        the process of pyre file monitor associated to it is killed.
    """

    NAME = "file_monitor"

    def __init__(
        self,
        configuration: Configuration,
        current_directory: str,
        analysis_directory: AnalysisDirectory,
    ) -> None:
        base_path = os.path.join(configuration.log_directory, self._name)
        super(ProjectFilesMonitor, self).__init__(base_path)
        self._configuration = configuration
        self._analysis_directory = analysis_directory

        self._extensions = set(
            ["py", "pyi", "thrift"] + configuration.extensions
        )  # type: Set[str]

        self._watchman_path = self._find_watchman_path(current_directory)  # type: str

        self.socket_connection = SocketConnection(self._configuration.log_directory)
        self.socket_connection.connect()
        self.socket_connection.perform_handshake(self._configuration.version_hash)

    @property
    def _name(self) -> str:
        return self.NAME

    @property
    @functools.lru_cache(1)
    def _subscriptions(self) -> List[Subscription]:
        subscription = {
            "expression": [
                "allof",
                ["type", "f"],
                ["not", "empty"],
                [
                    "anyof",
                    *[["suffix", extension] for extension in self._extensions],
                    ["match", "TARGETS"],
                ],
            ],
            "fields": ["name"],
        }
        return [
            Subscription(
                self._watchman_path, "pyre_file_change_subscription", subscription
            )
        ]

    @staticmethod
    def pid_path(configuration: Configuration) -> str:
        return os.path.join(
            configuration.log_directory,
            ProjectFilesMonitor.NAME,
            "{}.pid".format(ProjectFilesMonitor.NAME),
        )

    @staticmethod
    def is_alive(configuration: Configuration) -> bool:
        pid_path = ProjectFilesMonitor.pid_path(configuration)
        try:
            with open(pid_path) as file:
                pid = int(file.read())
                os.kill(pid, 0)  # throws if process is not running
            return True
        except Exception:
            return False

    def _handle_response(self, response: Dict[str, Any]) -> None:
        try:
            absolute_paths = [
                os.path.join(response["root"], path) for path in response["files"]
            ]
            LOG.info("Received Watchman update for files %s.", absolute_paths)

            updated_paths = self._analysis_directory.process_updated_files(
                absolute_paths
            )

            if updated_paths.is_empty():
                LOG.info("Skipping update: Pyre doesn't track any of these files.")
                return

            LOG.info(
                "Notifying server of update to files %s and invalidation of %s.",
                updated_paths.updated_paths,
                updated_paths.deleted_paths,
            )
            message = json_rpc.Request(
                method="updateFiles",
                parameters={
                    "files": updated_paths.updated_paths,
                    "invalidated": updated_paths.deleted_paths,
                },
            )
            if not message.write(self.socket_connection.output):
                LOG.info("Failed to communicate with server. Shutting down.")
                self._alive = False  # terminate daemon
                self.socket_connection.close()

        except KeyError:
            pass

    @staticmethod
    def _find_watchman_path(directory: str) -> str:
        watchman_path = find_root(directory, ".watchmanconfig")
        if not watchman_path:
            raise MonitorException(
                "Could not find a watchman directory from "
                "the current directory `{}`".format(directory)
            )
        return watchman_path

    @staticmethod
    def stop_project_monitor(configuration: Configuration) -> None:
        try:
            pid_path = ProjectFilesMonitor.pid_path(configuration)
            with open(pid_path) as file:
                pid = int(file.read())
                os.kill(pid, 2)  # sigint
                LOG.info("Stopped the file monitor.")
        except (FileNotFoundError, OSError, ValueError):
            pass



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import argparse
import os
from typing import List, Optional  # noqa

from .. import assert_writable_directory, log, readable_directory
from ..analysis_directory import AnalysisDirectory, resolve_analysis_directory
from ..configuration import Configuration
from .check import Check


def writable_directory(path: str) -> str:
    # Create the directory if it does not exist.
    try:
        os.makedirs(path)
    except FileExistsError:
        pass
    assert_writable_directory(path)
    return path


class Analyze(Check):
    NAME = "analyze"  # type: str

    def __init__(
        self,
        arguments: argparse.Namespace,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super(Analyze, self).__init__(
            arguments, original_directory, configuration, analysis_directory
        )
        self._analysis = arguments.analysis  # type: str
        self._taint_models_path = (
            arguments.taint_models_path or self._configuration.taint_models_path
        )  # type: List[str]
        self._no_verify = arguments.no_verify  # type: bool
        self._save_results_to = arguments.save_results_to  # type: Optional[str]
        self._dump_call_graph = arguments.dump_call_graph  # type: bool
        self._repository_root = arguments.repository_root  # type: Optional[str]
        self._rules = arguments.rule  # type: Optional[List[int]]

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        analyze = parser.add_parser(cls.NAME)
        analyze.set_defaults(command=cls)
        analyze.add_argument(
            "analysis",
            nargs="?",
            default="taint",
            help="Type of analysis to run: {taint}",
        )
        analyze.add_argument(
            "--taint-models-path",
            action="append",
            default=[],
            type=readable_directory,
            help="Location of taint models",
        )
        analyze.add_argument(
            "--no-verify",
            action="store_true",
            help="Do not verify models for the taint analysis.",
        )
        analyze.add_argument(
            "--save-results-to",
            default=None,
            type=writable_directory,
            help="Directory to write analysis results to.",
        )
        analyze.add_argument("--dump-call-graph", action="store_true")
        analyze.add_argument("--repository-root", type=os.path.abspath)
        analyze.add_argument("--rule", action="append", type=int)

    def generate_analysis_directory(self) -> AnalysisDirectory:
        return resolve_analysis_directory(
            self._arguments,
            self._configuration,
            self._original_directory,
            self._current_directory,
            build=True,
            isolate=True,
        )

    def _flags(self) -> List[str]:
        flags = super()._flags()
        flags.extend(["-analysis", self._analysis])
        if self._taint_models_path:
            for path in self._taint_models_path:
                flags.extend(["-taint-models", path])
        save_results_to = self._save_results_to
        if save_results_to:
            flags.extend(["-save-results-to", save_results_to])
        if self._dump_call_graph:
            flags.append("-dump-call-graph")
        if self._no_verify:
            flags.append("-no-verify")
        repository_root = self._repository_root
        if repository_root:
            flags.extend(["-repository-root", repository_root])
        rules = self._rules
        if rules is not None:
            flags.extend(["-rules", ",".join(str(rule) for rule in rules)])
        return flags

    def _run(self, retries: int = 1) -> None:
        self._analysis_directory.prepare()
        result = self._call_client(command=self.NAME)
        result.check()
        log.stdout.write(result.output)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import itertools
import json
import sys
from typing import Any, Dict, List, Optional, Tuple

from .postprocess import LOG


def json_to_errors(json_string: Optional[str]) -> List[Dict[str, Any]]:
    if json_string:
        try:
            return json.loads(json_string)
        except json.decoder.JSONDecodeError:
            LOG.error(
                "Recevied invalid JSON as input."
                "If piping from `pyre check` be sure to use `--output=json`."
            )
    else:
        LOG.error(
            "Recevied no input."
            "If piping from `pyre check` be sure to use `--output=json`."
        )
    return []


def sort_errors(errors: List[Dict[str, Any]]) -> List[Tuple[str, List[Any]]]:
    def error_path(error):
        return error["path"]

    return itertools.groupby(sorted(errors, key=error_path), error_path)


def filter_errors(arguments, errors: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    def matches_error_code(error) -> bool:
        return error["code"] == arguments.only_fix_error_code

    if arguments.only_fix_error_code:
        errors = list(filter(matches_error_code, errors))
    return errors


def errors_from_stdin(_arguments) -> List[Dict[str, Any]]:
    input = sys.stdin.read()
    errors = json_to_errors(input)
    return filter_errors(_arguments, errors)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import json
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple

from ..analysis_directory import AnalysisDirectory
from ..configuration import Configuration
from .command import Command, ProfileOutput


LOG: logging.Logger = logging.getLogger(__name__)

PHASE_NAME: str = "phase_name"


@dataclass(frozen=True)
class EventMetadata:
    name: str
    pid: int
    timestamp: int
    tags: Dict[str, str]


@dataclass(frozen=True)
class Event:
    metadata: EventMetadata

    # pyre-fixme[13]: Attribute `metadata` is never initialized.
    def __init__(self, metadata: EventMetadata) -> None:
        raise NotImplementedError


@dataclass(frozen=True)
class DurationEvent(Event):
    duration: int

    def add_phase_duration_to_result(self, result: Dict[str, int]) -> None:
        tags = self.metadata.tags
        if PHASE_NAME in tags:
            phase_name = tags[PHASE_NAME]
            result[phase_name] = self.duration


@dataclass(frozen=True)
class CounterEvent(Event):
    description: Optional[str]


def _parse_tags(input: List[List[str]]) -> Dict[str, str]:
    return {key: value for [key, value] in input}


def _parse_metadata(input_json: Dict[str, Any]) -> EventMetadata:
    return EventMetadata(
        name=input_json["name"],
        pid=input_json["pid"],
        timestamp=input_json["timestamp"],
        tags=_parse_tags(input_json.get("tags", [])),
    )


def parse_event(input_string: str) -> Event:
    input_json: Dict[str, Any] = json.loads(input_string)
    event_type = input_json["event_type"]
    metadata = _parse_metadata(input_json)
    if event_type[0] == "Duration":
        duration = event_type[1]
        return DurationEvent(duration=duration, metadata=metadata)
    elif event_type[0] == "Counter":
        description = None if len(event_type) <= 1 else event_type[1]
        return CounterEvent(description=description, metadata=metadata)
    else:
        raise ValueError("Unrecognized event type: {}".format(input))


def parse_events(input_string: str) -> List[Event]:
    output: List[Event] = []
    for index, line in enumerate(input_string.splitlines()):
        try:
            line = line.strip()
            if len(line) == 0:
                continue
            output.append(parse_event(line))
        except Exception:
            raise RuntimeError(
                "Malformed log entry detected on line {}".format(index + 1)
            )
    return output


def to_traceevents(events: Sequence[Event]) -> List[Dict[str, Any]]:
    def to_traceevent(event: Event) -> Optional[Dict[str, Any]]:
        if isinstance(event, DurationEvent):
            duration_ms = event.duration
            start_time_ms = event.metadata.timestamp - duration_ms
            return {
                "pid": event.metadata.pid,
                "tid": 0,
                "ts": start_time_ms * 1000,
                "ph": "X",
                "name": event.metadata.name,
                "dur": duration_ms * 1000,
                "args": event.metadata.tags,
            }
        elif isinstance(event, CounterEvent):
            timestamp_ms = event.metadata.timestamp
            arguments: Dict[str, Any] = {
                key: int(value) for key, value in event.metadata.tags.items()
            }
            return {
                "pid": event.metadata.pid,
                "tid": 0,
                "ts": timestamp_ms * 1000,
                "ph": "C",
                "name": event.metadata.name,
                "args": arguments,
            }
        else:
            return None

    return [
        trace_event
        for trace_event in map(to_traceevent, events)
        if trace_event is not None
    ]


def split_pre_and_post_initialization(
    events: Sequence[Event]
) -> Tuple[Sequence[Event], Sequence[Event]]:
    initialization_point = next(
        (
            index
            for index, event in enumerate(events)
            if event.metadata.name == "initialization"
        ),
        len(events) - 1,
    )
    return events[:initialization_point], events[initialization_point:]


def to_cold_start_phases(events: Sequence[Event]) -> Dict[str, int]:
    result: Dict[str, int] = {}
    pre_initialization_events, _ = split_pre_and_post_initialization(events)
    for event in pre_initialization_events:
        if not isinstance(event, DurationEvent):
            continue
        event.add_phase_duration_to_result(result)

    return result


def to_incremental_updates(events: Sequence[Event]) -> List[Dict[str, int]]:
    results: List[Dict[str, int]] = []
    current: Dict[str, int] = {}
    _, post_initialization_events = split_pre_and_post_initialization(events)
    for event in post_initialization_events:
        if not isinstance(event, DurationEvent):
            continue

        event.add_phase_duration_to_result(current)

        if event.metadata.name == "incremental check":
            current["total"] = event.duration
            results.append(current)
            current = {}
    return results


class Profile(Command):
    NAME = "profile"

    def __init__(
        self,
        arguments: argparse.Namespace,
        original_directory: str,
        configuration: Optional[Configuration] = None,
        analysis_directory: Optional[AnalysisDirectory] = None,
    ) -> None:
        super(Profile, self).__init__(
            arguments, original_directory, configuration, analysis_directory
        )
        self._profile_output: ProfileOutput = arguments.profile_output

    @classmethod
    def add_subparser(cls, parser: argparse._SubParsersAction) -> None:
        profile = parser.add_parser(cls.NAME)
        profile.set_defaults(command=cls)
        profile.add_argument(
            "--profile-output",
            type=ProfileOutput,
            choices=ProfileOutput,
            help="Specify what to output.",
            default=ProfileOutput.COLD_START_PHASES,
        )

    def _run(self) -> None:
        try:
            profiling_output = Path(self.profiling_log_path())
            if not profiling_output.is_file():
                raise RuntimeError(
                    "Cannot find profiling output at `{}`. "
                    "Please run Pyre with `--enable-profiling` or "
                    "`--enable-memory-profiling` option first.".format(profiling_output)
                )
            events = parse_events(profiling_output.read_text())
            output = self._profile_output
            if output == ProfileOutput.TRACE_EVENT:
                print(json.dumps(to_traceevents(events)))
            elif output == ProfileOutput.COLD_START_PHASES:
                print(json.dumps(to_cold_start_phases(events), indent=2))
            elif output == ProfileOutput.INCREMENTAL_UPDATES:
                print(json.dumps(to_incremental_updates(events), indent=2))
            else:
                raise RuntimeError("Unrecognized output format: {}".format(output))

        except Exception as e:
            LOG.error("Failed to inspect profiling log: {}".format(e))
            raise e
